# Copyright (C) 2006-2017 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import logging
import os
import re
import subprocess
import stat
import tempfile
import time

from agent.extension import OVSDiscoverExtension
from agent.lib.config import get_config
from agent.lib.settings import (get_ovn_zookeeper_certificates,
                                write_ovn_zookeeper_certificates,
                                remove_ovn_zookeeper_certificates,
                                get_ovn_conf,
                                write_ovn_conf,
                                remove_ovn_conf,
                                remove_ovn_conf_dir,
                                get_ovn_server_certificate)


OVM_SYS_REPO_PART_PREFIX = 'OVM_SYS_REPO_PART_'


def get_ovs_info():
    ovs_info = {}
    f = open("/etc/ovs-info")
    for line in f:
        m = re.match(r'^release:(.*)$', line)
        if m:
            ovs_info['release'] = m.group(1).strip()
        m = re.match(r'^date:(.*)$', line)
        if m:
            ovs_info['date'] = m.group(1).strip()
        m = re.match(r'^build:(.*)$', line)
        if m:
            ovs_info['build'] = m.group(1).strip()
    f.close()
    return ovs_info


def run_cmd(cmd, success_return_code=(0,)):
    if not isinstance(cmd, list):
        raise Exception("Only accepts list!")
    cmd = [str(x) for x in cmd]
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, close_fds=True)
    (stdoutdata, stderrdata) = proc.communicate()
    if proc.returncode not in success_return_code:
        raise RuntimeError('Command: %s failed (%s): stderr: %s stdout: %s'
                           % (cmd, proc.returncode, stderrdata, stdoutdata))
    return str(stdoutdata)


def retry_cmd(cmd, interval=5, count=3):
    # retry if failed
    for i in range(1, count):
        try:
            return run_cmd(cmd)
        except Exception, e:
            logging.debug("retry_cmd: command %s failed, error %s"
                          "retrying in %s seconds", cmd, e, interval)
            time.sleep(interval)
    # retry for the last time
    return run_cmd(cmd)


def get_server_unique_id():
    """
    Obtain a globally unique identifier for this server.

    This invokes legacy C code from the VI agent and returns a 16 byte
    unique ID in the form of a string of hex digits:
        xx:xx:xx:xx:xx:xx:xx:xx:xx:xx:xx:xx:xx:xx:xx:xx

    @return    The UUID string or None on error
    """
    try:
        config = get_config()
        fakeuuid = config.get("server", "fakeuuid")
    except Exception:
        fakeuuid = ""

    if fakeuuid != "":
        return fakeuuid

    try:
        status, response, logmsg = OVSDiscoverExtension.unique_id(None)
        if status == 0:
            return response
        else:
            logging.error(logmsg)
    except Exception, err:
        logging.error('Error getting server unique id: %s', err)


def set_nice(increment=10, io_cls=2, io_data=7):
    os.nice(increment)
    run_cmd(['ionice', '-c', io_cls, '-n', io_data, '-p', os.getpid()])


def reset_console_access():
    run_cmd(['service', 'ovm-consoled', 'restart'])


def guess_mimetype(filename):
    ftype = run_cmd(['file', '-b', '-i', filename])
    ftype = ftype.split(',')[0] # application/x-tar, POSIX
    ftype = ftype.split(';')[0] # application/x-tar; charset=binary
    return ftype.strip()


def host_alive(ip):
    status = os.popen("ping -q -c 2 %s" % ip).close()
    return not status


def get_netmask(ip_address):
    """
    Return the netmask of the network where the specified IP address
    is defined. The IP address should be defined on a network interface
    of the server, otherwise None is returned.

    """
    output = run_cmd(["ifconfig"]).splitlines()
    for line in output:
        if (':' + ip_address + ' ') in line:
            netmask = line.split(':')[3]
            return netmask


def _get_ip_dev(ip):
    """Get dev the ip is bound to."""
    cmd = ["ip", "addr", "show", "to", ip]
    output = run_cmd(cmd).splitlines()
    for line in output:
        if "scope global" in line:
            return line.split()[-1]
    return ''


def iscsi_logout_target(target, portal):
    cmd = ["/sbin/iscsiadm",
           "--mode", "node",
           "--targetname=%s" % target,
           "--portal=%s" % portal,
           "--logout"]
    run_cmd(cmd)


def iscsi_delete_portal(portal):
    cmd = ["/sbin/iscsiadm",
           "--mode", "node",
           "--portal=%s" % portal,
           "--op=delete"],
    run_cmd(cmd)


def get_file_size(filename):
    """
    Return the size of a file. If the file is a block device then
    this returns the size of the block device.

    """
    file_stat = os.stat(filename)
    file_size = 0
    if stat.S_ISREG(file_stat[stat.ST_MODE]):
        file_size = file_stat[stat.ST_SIZE]
    elif stat.S_ISBLK(file_stat[stat.ST_MODE]):
        file_size = run_cmd(['/sbin/blockdev', '--getsize64', filename])
    return long(file_size)


def disk_cfg_get_device(disk_cfg):
    #
    # This function returns 3 items:
    # 1) the actual path of the disk
    # 2) the guid of the disk
    # 3) the name of the guid, to be used when creating a file using
    # guid.  On X86 Linux, the guid and the name is actually the same.
    #
    src = disk_cfg.split(':')[-1]
    if not src:
        return (None, None, None)
    file_id = os.path.splitext(os.path.basename(src))[0]
    return (src, file_id, file_id)


def get_copy_cmd(src, dst, sparse):
    """
    Return the appropriate command to copy src to dst.

    """
    block_sz = '1M'
    min_file_sz = 1073741824

    try:
        src_size = get_file_size(src)
        if src_size >= min_file_sz:
            block_sz = '8M'
    except Exception, e:
        logging.debug("get_copy_cmd: failed to get file size, error %s", e)

    logging.debug("get_copy_cmd: block size %s", block_sz)

    args = ['/bin/dd', 'if=%s' % src, 'of=%s' % dst, 'bs=%s' % block_sz,
            'iflag=direct', 'oflag=direct']
    if sparse:
        args.append('conv=sparse')
    return args


def sync():
    """
    Synchronize data on disk with memory

    """
    run_cmd(['sync'])


def get_fs_type(fs_location):
    if fs_location.startswith('/dev/'):
        return 'ocfs2'
    else:
        return 'nfs'


def do_export_nfs(client, path, options):
    if not os.path.ismount(path):
        logging.warn('%s is not a mount point', path)
    if not options:
        options = 'rw,async,no_root_squash'
    cmd = ['service', 'nfs', 'start']
    run_cmd(cmd)
    # First un-export to remove any existing entries for this client/path pair
    # in export file /var/lib/nfs/etab
    client_path = '%s:%s' % (client, path)
    client_path_stripped = client_path.rstrip('/')
    cmd = ['exportfs', '-u', client_path_stripped]
    run_cmd(cmd)
    cmd = ['exportfs', '-o', options, client_path]
    run_cmd(cmd)


def do_unexport_nfs(client, path):
    client_path = '%s:%s' % (client, path.rstrip('/'))
    cmd = ['exportfs', '-u', client_path]
    run_cmd(cmd)
    if not run_cmd(['exportfs']):
        cmd = ['service', 'nfs', 'stop']
        run_cmd(cmd)


def mount(resource, mountpoint, fstype=None):
    """
    Mount a file system.

    Mount a file system. The mount point is created if it does not
    exist. This does nothing if the file system is already mounted.
    If the mount fails then the mount point is removed and an
    exception is raised.

    @param resource: The resource to mount.
    @type  resource: C{string}

    @param mountpoint: The mount point where the resource has to be
    mounted.
    @type  mountpoint: C{string}

    @param fstype: The type of file system to mount.
    @type  fstype: C{string}

    """
    if os.path.ismount(mountpoint):
        return

    if not os.path.exists(mountpoint):
        os.makedirs(mountpoint)

    cmd = ['mount']
    if fstype:
        cmd += ['-t', fstype]

    cmd += [resource, mountpoint]

    try:
        run_cmd(cmd)
    except Exception, err:
        try:
            os.rmdir(mountpoint)
        except Exception, e:
            logging.warn('Error removing mount point: %s', mountpoint)
        raise err


def umount(mountpoint):
    """
    Unmount a file system.

    Unmount a file system. The mount point is removed after the file
    system has been successfully unmounted. This does nothing if the
    file system is not mounted.

    @param mountpoint: The mount point that should be unmounted.
    @type  mountpoint: C{string}

    """
    if not os.path.ismount(mountpoint):
        return

    run_cmd(['umount', mountpoint])
    os.rmdir(mountpoint)


def get_mounts():
    mounts = {}
    mount_output = run_cmd(['mount'])
    for entry in mount_output.splitlines():
        fields = entry.split()
        dev = fields[0]
        mntpoint = fields[2]
        mounts[mntpoint] = dev
    return mounts


def get_swaps():
    swaps = []
    swaps_file = open('/proc/swaps')
    for entry in swaps_file:
        fields = entry.split()
        filename = fields[0]
        if filename == 'Filename':
            continue
        swaps.append(filename)
    swaps_file.close()
    return swaps


def get_rootdev():
    mounts = get_mounts()
    rootdev = mounts.get('/', '')
    if 'md' not in rootdev:
        rootdev = rootdev.rstrip('0123456789')
    if 'cciss' in rootdev:
        rootdev = rootdev[:-1] # remove trailing p
    return rootdev


def get_dev_partitions(device):
    partitions = []
    if device:
        try:
            output = run_cmd(['lsblk', device, '-l', '-o', 'NAME,TYPE'])
            for entry in output.splitlines():
                fields = entry.split()
                if fields[1] == "part":
                    partitions.append('/dev/%s' % fields[0])
        except Exception, err:
            logging.error('Error getting dev partitions: %s', err)
    return partitions


def get_bootdev():
    mounts = get_mounts()
    bootdev = mounts.get('/boot', '')
    if 'md' not in bootdev:
        bootdev = bootdev.rstrip('0123456789')
    return bootdev


def read_file(*args):
    path = os.path.join(*args)
    if os.path.exists(path):
        f = open(path)
        content = f.read()
        f.close()
        return content.strip()
    else:
        return ''


def get_dev_info(dev):
    info = {}
    if dev.startswith('/dev/'):
        dev = dev[5:]
    if 'cciss' in dev:
        dev_sys = dev.replace('/', '!')
        try:
            serial = run_cmd(['cciss_id', '/dev/%s' % dev])
            serial = serial.strip()
        except Exception, err:
            logging.debug('Error getting %s cciss id: %s', dev, err)
            serial = read_file('/sys/block', dev_sys, 'device/unique_id')
            serial = serial.lower().replace(' ', '_')
        if serial:
            info['ID_SERIAL'] = serial.strip()
        vendor = read_file('/sys/block', dev_sys, 'device/vendor')
        if vendor:
            info['ID_VENDOR'] = vendor
        model = read_file('/sys/block', dev_sys, 'device/model')
        if model:
            info['ID_MODEL'] = model
        revision = read_file('/sys/block', dev_sys, 'device/rev')
        if revision:
            info['ID_REVISION'] = revision
    else:
        output = run_cmd(['scsi_id', '-g', '-x', '-u', '-d', '/dev/%s' % dev])
        for entry in output.splitlines():
            key, value = entry.split('=', 1)
            info[key] = value
    return info


def get_root_dev_info():
    info = {}
    try:
        rootdev = get_rootdev()
        if rootdev and 'md' not in rootdev: # ignore MD Raid device
            info = get_dev_info(rootdev)
    except Exception, err:
        logging.error('Error getting root dev info: %s', err)
    return info


def get_repository_partition():
    device = ''
    rootdev = get_rootdev()
    if not rootdev:
        return device
    rootdev_partitions = get_dev_partitions(rootdev)
    if not rootdev_partitions:
        return device
    mounts = get_mounts()
    swaps = get_swaps()
    partitions = open('/proc/partitions')
    for entry in partitions:
        fields = entry.split()
        if len(fields) != 4:
            continue
        dev_blocks = fields[2]
        if dev_blocks == '#blocks': # ignore heading
            continue
        if int(dev_blocks) < 2097152: # ignore partitions < 2G
            continue
        dev_name = fields[3]
        dev_path = '/dev/%s' % dev_name
        if dev_path == rootdev:
            continue
        if dev_path in swaps:
            continue
        if dev_path in rootdev_partitions and dev_path not in mounts.values():
            device = dev_path
            break
    partitions.close()
    return device


def init_repository_partition():
    device = get_repository_partition()
    if not device:
        return

    dev_info = get_root_dev_info()
    dev_id = dev_info.get('ID_SERIAL')
    if not dev_id:
        return

    dm_name = OVM_SYS_REPO_PART_PREFIX + dev_id

    dm_dev = '/dev/mapper/%s' % dm_name
    if os.path.exists(dm_dev):
        return

    try:
        device_secs = run_cmd(['blockdev', '--getsz', device]).strip()
    except Exception, err:
        logging.warn('Error getting device %s size: %s', device, err)
        return
    dm_table = '0 %s linear %s 0' % (device_secs, device)
    try:
        run_cmd(['dmsetup', 'create', dm_name, '--table', dm_table, '-u', dm_name])
    except Exception, err:
        logging.error('Error init repository partition: %s', err)


def get_rpm_version(package):
    return run_cmd(['rpm', '-q', '--queryformat', '%{VERSION}-%{RELEASE}', package])


def rpm_install(rpm, options=None):
    fd, fn = tempfile.mkstemp(suffix='.rpm', prefix="rpm-")
    try:
        os.write(fd, rpm.data)
        os.close(fd)
        cmd = ['rpm', '--upgrade']
        if options:
            cmd += options
        cmd += [fn]
        run_cmd(cmd)
    finally:
        os.unlink(fn)


def rpm_uninstall(rpms, options=None):
    cmd = ['rpm', '--erase']
    if options:
        cmd += options
    cmd += rpms
    run_cmd(cmd)


def service_enable(service):
    run_cmd(['chkconfig', '--level', '2345', service, 'on'])


def service_disable(service):
    run_cmd(['chkconfig', '--level', '2345', service, 'off'])


def service_start(service):
    run_cmd(['service', service, 'start'])


def service_stop(service):
    run_cmd(['service', service, 'stop'])


def get_service_status(service):
    try:
        run_cmd(['service', service, 'status'])
        return 'RUNNING'
    except Exception:
        return 'NOTRUNNING'


def ovn_server_gen_cert():
    run_cmd(['ovn-server-gen-cert.sh'])


def ovn_configure(config):
    ovn_conf = config.get('ovn_conf')
    if not ovn_conf:
        raise Exception('ovn_conf is required.')
    zookeeper_certificates = config.get('zookeeper_certificates')
    if not zookeeper_certificates:
        raise Exception('zookeeper_certificates is required.')
    try:
        get_rpm_version('ovna')
    except Exception:
        raise Exception('ovna is not installed.')
    write_ovn_conf(ovn_conf)
    write_ovn_zookeeper_certificates(zookeeper_certificates)
    ovn_server_gen_cert()


def ovn_unconfigure():
    remove_ovn_zookeeper_certificates()
    remove_ovn_conf()
    remove_ovn_conf_dir()


def ovn_status():
    status = {}
    status['ovn_conf'] = get_ovn_conf()
    status['ovn_zookeeper_certificates'] = get_ovn_zookeeper_certificates()
    status['ovn_server_certificate'] = get_ovn_server_certificate()
    status['ovn_ovna_service'] = get_service_status('ovna')
    status['ovn_openvswitch_service'] = get_service_status('openvswitch')
    return status
